/****************************************************************************
 ****************************************************************************
 * Palm Wireless Keyboard(TM) infrared signal analyzer for linux. v0.1.0
 * Copyright GPL Licence. I quoted some part of source code from
 * mode2.c in the lirc-0.6.6.tar.gz package. Its Copyright is bellow.
 * Copyright (C) 2004. 12. 25. lampo1 <lampo1@paran.com>
 *
 ****************************************************************************
 * mode2 - shows the pulse/space length of a remote button
 *
 * Copyright (C) 1998 Trent Piepho <xyzzy@u.washington.edu>
 * Copyright (C) 1998 Christoph Bartelmus <lirc@bartelmus.de>
 *
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include "lirc.h"

#define MAX_SCODE  3
#define MAX_KCODE  3
#define MAX_NAME   14
#define MAX_IRSIG  60

struct pwkey {
    int    serial;
    char   pwkname[MAX_NAME];
    char   nml_xname[MAX_NAME];
    char   lfn_xname[MAX_NAME];
    char   rfn_xname[MAX_NAME];
    int    irsig[MAX_IRSIG];
    int    nml_scode[MAX_SCODE];
    int    lfn_scode[MAX_SCODE];
    int    rfn_scode[MAX_SCODE];
    int    nml_kcode[MAX_KCODE];
    int    lfn_kcode[MAX_KCODE];
    int    rfn_kcode[MAX_KCODE];
};

struct pwkey key[103] = {
    {0,"REPEAT","none","none","none",{},{},{},{},{},{},{}},
    {1,"LFN_mk","none","none","none",{},{},{},{},{},{},{}},
    {2,"LFN_bk","none","none","none",{},{},{},{},{},{},{}},    
    {3,"RFN_mk","none","none","none",{},{},{},{},{},{},{}},
    {4,"RFN_bk","none","none","none",{},{},{},{},{},{},{}},
    {5,"CMD_mk","Escape","none","none",{},{0x01,},{},{},{},{},{}},
    {6,"CMD_bk","Escape","none","none",{},{0x81,},{},{},{},{},{}},
    {7,"A_mk","a","none","none",{},{0x1E,},{},{},{},{},{}},
    {8,"A_bk","a","none","none",{},{0x9E,},{},{},{},{},{}},
    {9,"B_mk","b","none","none",{},{0x30,},{},{},{},{},{}},
    {10,"B_bk","b","none","none",{},{0xB0,},{},{},{},{},{}},
    {11,"C_mk","c","none","none",{},{0x2E,},{},{},{},{},{}},
    {12,"C_bk","c","none","none",{},{0xAE,},{},{},{},{},{}},
    {13,"D_mk","d","none","none",{},{0x20,},{},{},{},{},{}},
    {14,"D_bk","d","none","none",{},{0xA0,},{},{},{},{},{}},
    {15,"E_mk","e","3","numbersign",{},{0x12,},{0x04,},{0x2A,0x04,},{},{},{}},
    {16,"E_bk","e","3","numbersign",{},{0x92,},{0x84,},{0x84,0xAA,},{},{},{}},
    {17,"F_mk","f","none","none",{},{0x21,},{},{},{},{},{}},
    {18,"F_bk","f","none","none",{},{0xA1,},{},{},{},{},{}},
    {19,"G_mk","g","none","none",{},{0x22,},{},{},{},{},{}},
    {20,"G_bk","g","none","none",{},{0xA2,},{},{},{},{},{}},
    {21,"H_mk","h","none","none",{},{0x23,},{},{},{},{},{}}, 
    {22,"H_bk","h","none","none",{},{0xA3,},{},{},{},{},{}},
    {23,"I_mk","i","8","asterisk",{},{0x17,},{0x09,},{0x2A,0x09,},{},{},{}},
    {24,"I_bk","i","8","asterisk",{},{0x97,},{0x89,},{0x89,0xAA,},{},{},{}},
    {25,"J_mk","j","none","none",{},{0x24,},{},{},{},{},{}},
    {26,"J_bk","j","none","none",{},{0xA4,},{},{},{},{},{}},
    {27,"K_mk","k","none","none",{},{0x25,},{},{},{},{},{}},
    {28,"K_bk","k","none","none",{},{0xA5,},{},{},{},{},{}},
    {29,"L_mk","l","none","none",{},{0x26,},{},{},{},{},{}},
    {30,"L_bk","l","none","none",{},{0xA6,},{},{},{},{},{}},
    {31,"M_mk","m","none","none",{},{0x32,},{},{},{},{},{}},
    {32,"M_bk","m","none","none",{},{0xB2,},{},{},{},{},{}},
    {33,"N_mk","n","none","none",{},{0x31,},{},{},{},{},{}},
    {34,"N_bk","n","none","none",{},{0xB1,},{},{},{},{},{}},
    {35,"O_mk","o","9","parenleft",{},{0x18,},{0x0A,},{0x2A,0x0A,},{},{},{}},
    {36,"O_bk","o","9","parenleft",{},{0x98,},{0x8A,},{0x8A,0xAA,},{},{},{}},
    {37,"P_mk","p","0","parenright",{},{0x19,},{0x0B,},{0x2A,0x0B,},{},{},{}},
    {38,"P_bk","p","0","parenright",{},{0x99,},{0x8B,},{0x8B,0xAA,},{},{},{}},
    {39,"Q_mk","q","1","exclam",{},{0x10,},{0x02,},{0x2A,0x02,},{},{},{}},
    {40,"Q_bk","q","1","exclam",{},{0x90,},{0x82,},{0x82,0xAA,},{},{},{}},
    {41,"R_mk","r","4","dollar",{},{0x13,},{0x05,},{0x2A,0x05,},{},{},{}},
    {42,"R_bk","r","4","dollar",{},{0x93,},{0x85,},{0x85,0xAA,},{},{},{}},
    {43,"S_mk","s","none","none",{},{0x1F,},{},{},{},{},{}},
    {44,"S_bk","s","none","none",{},{0x9F,},{},{},{},{},{}},
    {45,"T_mk","t","5","percent",{},{0x14,},{0x06,},{0x2A,0x06,},{},{},{}},
    {46,"T_bk","t","5","percent",{},{0x94,},{0x86,},{0x86,0xAA,},{},{},{}},
    {47,"U_mk","u","7","ampersand",{},{0x16,},{0x08,},{0x2A,0x08,},{},{},{}},
    {48,"U_bk","u","7","ampersand",{},{0x96,},{0x88,},{0x88,0xAA,},{},{},{}},
    {49,"V_mk","v","none","none",{},{0x2F,},{},{},{},{},{}},
    {50,"V_bk","v","none","none",{},{0xAF,},{},{},{},{},{}},
    {51,"W_mk","w","2","at",{},{0x11,},{0x03,},{0x2A,0x03,},{},{},{}},
    {52,"W_bk","w","2","at",{},{0x91,},{0x83,},{0x83,0xAA,},{},{},{}},
    {53,"X_mk","x","none","none",{},{0x2D,},{},{},{},{},{}},
    {54,"X_bk","x","none","none",{},{0xAD,},{},{},{},{},{}},
    {55,"Y_mk","y","6","asciicircum",{},{0x15,},{0x07,},{0x2A,0x07,},{},{},{}},
    {56,"Y_bk","y","6","asciicircum",{},{0x95,},{0x87,},{0x87,0xAA,},{},{},{}},
    {57,"Z_mk","z","none","none",{},{0x2C,},{},{},{},{},{}},
    {58,"Z_bk","z","none","none",{},{0xAC,},{},{},{},{},{}},
    {59,"MINUS_mk","minus","bracketleft","braceleft",{},{0x0C,},{0x1A,},{0x2A,0x1A,},{},{},{}},
    {60,"MINUS_bk","minus","bracketleft","braceleft",{},{0x8C,},{0x9A,},{0x9A,0xAA,},{},{},{}},
    {61,"EQUAL_mk","equal","bracketright","braceright",{},{0x0D,},{0x1B,},{0x2A,0x1B,},{},{},{}},
    {62,"EQUAL_bk","equal","bracketright","braceright",{},{0x8D,},{0x9B,},{0x9B,0xAA,},{},{},{}},
    {63,"BKSP_mk","BackSpace","none","none",{},{0x0E,},{},{},{},{},{}},
    {64,"BKSP_bk","BackSpace","none","none",{},{0x8E,},{},{},{},{},{}},
    {65,"LSPACE_mk","space","none","none",{},{0x39,},{},{},{},{},{}},
    {66,"LSPACE_bk","space","none","none",{},{0xB9,},{},{},{},{},{}},
    {67,"RSPACE_mk","space","none","none",{},{0x39,},{},{},{},{},{}},
    {68,"RSPACE_bk","space","none","none",{},{0xB9,},{},{},{},{},{}},
    {69,"TAB_mk","Tab","none","none",{},{0x0F,},{},{},{},{},{}},
    {70,"TAB_bk","Tab","none","none",{},{0x8F,},{},{},{},{},{}},
    {71,"CAPS_mk","Caps_Lock","Num_Lock","none",{},{0x3A,},{0x45,},{},{},{},{}},
    {72,"CAPS_bk","Caps_Lock","Num_Lock","none",{},{0xBA,},{0xC5,},{},{},{},{}},
    {73,"LSHIFT_mk","Shift_L","none","none",{},{0x2A,},{},{},{},{},{}},
    {74,"LSHIFT_bk","Shift_L","none","none",{},{0xAA,},{},{},{},{},{}},
    {75,"RSHIFT_mk","Shift_R","none","none",{},{0x36,},{},{},{},{},{}},
    {76,"RSHIFT_bk","Shift_R","none","none",{},{0xB6,},{},{},{},{},{}},
    {77,"LCTRL_mk","Control_L","none","none",{},{0x1D,},{},{},{},{},{}},
    {78,"LCTRL_bk","Control_L","none","none",{},{0x9D,},{},{},{},{},{}},
    {79,"LALT_mk","Alt_L","none","none",{},{0x38,},{},{},{},{},{}},
    {80,"LALT_bk","Alt_L","none","none",{},{0xB8,},{},{},{},{},{}},
    {81,"DELETE_mk","Delete","none","none",{},{0xE0,0x53,},{},{},{},{},{}},
    {82,"DELETE_bk","Delete","none","none",{},{0xE0,0xD3,},{},{},{},{},{}},
    {83,"UARROW_mk","Up","none","Prior",{},{0xE0,0x48,},{},{0xE0,0x49,},{},{},{}},
    {84,"UARROW_bk","Up","none","Prior",{},{0xE0,0xC8,},{},{0xE0,0xC9,},{},{},{}},
    {85,"DARROW_mk","Down","none","Next",{},{0xE0,0x50,},{},{0xE0,0x51,},{},{},{}},
    {86,"DARROW_bk","Down","none","Next",{},{0xE0,0xD0,},{},{0xE0,0xD1,},{},{},{}},
    {87,"LARROW_mk","Left","none","Home",{},{0xE0,0x4B,},{},{0xE0,0x47,},{},{},{}},
    {88,"LARROW_bk","Left","none","Home",{},{0xE0,0xCB,},{},{0xE0,0x97,},{},{},{}},
    {89,"RARROW_mk","Right","none","End",{},{0xE0,0x4D,},{},{0xE0,0x4F,},{},{},{}},
    {90,"RARROW_bk","Right","none","End",{},{0xE0,0xCD,},{},{0xE0,0xCF,},{},{},{}},
    {91,"SEMICOLON_mk","semicolon","none","none",{},{0x27,},{},{},{},{},{}},
    {92,"SEMICOLON_bk","semicolon","none","none",{},{0xA7,},{},{},{},{},{}},
    {93,"APOSTROPHE_mk","apostrophe","grave","asciitilde",{},{0x28,},{0x29,},{0x2A,0x29,},{},{},{}},
    {94,"APOSTROPHE_bk","apostrophe","grave","asciitilde",{},{0xA8,},{0x89,},{0x89,0xAA,},{},{},{}},
    {95,"COMMA_mk","comma","none","none",{},{0x33,},{},{},{},{},{}},
    {96,"COMMA_bk","comma","none","none",{},{0xB3,},{},{},{},{},{}},
    {97,"PERIOD_mk","period","none","none",{},{0x34,},{},{},{},{},{}},
    {98,"PERIOD_bk","period","none","none",{},{0xB4,},{},{},{},{},{}},
    {99,"SLASH_mk","slash","backslash","bar",{},{0x35,},{0x2B,},{0x2A,0x2B,},{},{},{}},
    {100,"SLASH_bk","slash","backslash","bar",{},{0xB5,},{0xAB,},{0xAB,0xAA,},{},{},{}},
    {101,"ENTER_mk","Return","none","none",{},{0x1C,},{},{},{},{},{}},
    {102,"ENTER_bk","Return","none","none",{},{0x9C,},{},{},{},{},{}}
};

int   iter;               /* how many times to test per key */
int   max_err = 80;       /* limit of acceptable difference between signals */ 
int   start_idx = 0;      /* This variable is used in pwkdriver. From this  */
                          /* index, signals are compared.                   */      
     
int   irdata[220][MAX_IRSIG];
int   irsig[MAX_IRSIG];
int   sum_mk[MAX_IRSIG], sum_bk[MAX_IRSIG];
int   avr_mk[MAX_IRSIG], avr_bk[MAX_IRSIG];
int   high_mk[MAX_IRSIG], low_mk[MAX_IRSIG];
int   high_bk[MAX_IRSIG], low_bk[MAX_IRSIG];


int max(int a, int b) {
    if (a > b) return a;
    else return b;
}

int min(int a, int b) {
    if (a > b) return b;
    else return a;
}

void reset_irsig(int *sig) {
    int i;
    for (i = 0; i <= MAX_IRSIG-1; i++) sig[i] = 0;
}

void read_from_file(FILE *infd) {
    int i, j;
    char dmy[80];
        
    do {
	fgets(dmy, sizeof(dmy), infd);        
    } while( dmy[0] == '#' );

    sscanf(dmy, "%*s %*s %d %*s %*s %*s %d",&max_err, &start_idx);

    for (i = 0; i<= 102; i++) {
        fscanf(infd,"%d %s %s %s %s", &key[i].serial, key[i].pwkname,
  	   key[i].nml_xname, key[i].lfn_xname, key[i].rfn_xname);
        for(j=0; j <= MAX_IRSIG-1 ; j++) fscanf(infd,"%d:",&key[i].irsig[j]);
        for(j=0; j <= 2 ; j++) fscanf(infd,"%*x:");
        for(j=0; j <= 2 ; j++) fscanf(infd,"%*x:");
        for(j=0; j <= 2 ; j++) fscanf(infd,"%*x:");
        for(j=0; j <= 2 ; j++) fscanf(infd,"%d:",&key[i].nml_kcode[j]);
        for(j=0; j <= 2 ; j++) fscanf(infd,"%d:",&key[i].lfn_kcode[j]);
        for(j=0; j <= 2 ; j++) fscanf(infd,"%d:",&key[i].rfn_kcode[j]);
        fgets(dmy,sizeof(dmy),infd);
    }
    
    printf("max_err = %d, start_idx = %d\n", max_err, start_idx);
   
    for (i = 0; i <= 102; i++) {
        printf("%d %s %s %s %s ", i, key[i].pwkname, key[i].nml_xname,
      	   key[i].lfn_xname, key[i].rfn_xname);
        for(j=0; j <= MAX_IRSIG-1; j++) printf("%d:",key[i].irsig[j]);
        printf(" ");
        for(j=0; j <= 2 ; j++) printf("%x:",key[i].nml_scode[j]);
        printf(" ");
        for(j=0; j <= 2 ; j++) printf("%x:",key[i].lfn_scode[j]);
        printf(" ");
        for(j=0; j <= 2; j++)  printf("%x:",key[i].rfn_scode[j]);
        printf(" ");
        for(j=0; j <= 2 ; j++) printf("%d:",key[i].nml_kcode[j]);
        printf(" ");
        for(j=0; j <= 2 ; j++) printf("%d:",key[i].lfn_kcode[j]);
        printf(" ");
        for(j=0; j <= 2; j++)  printf("%d:",key[i].rfn_kcode[j]);
        printf("\n");
    }
}

void write_to_file(FILE *outfd) {
    int i,j;

    fprintf(outfd,"#This file contains IR data for Palm Wireless Keyboard.\n");
    fprintf(outfd,"#Field 7, 8, 9 are hex values.\n");
    fprintf(outfd,"#Field 10, 11, 12 are decimal values.\n");
    fprintf(outfd,"max_err = %d ; start_idx = %d\n", max_err, start_idx);
    for (i = 0; i <= 102; i++) {
        fprintf(outfd,"%d %s %s %s %s ", i, key[i].pwkname, key[i].nml_xname, 
		key[i].lfn_xname, key[i].rfn_xname);
        for(j=0; j <= MAX_IRSIG-1; j++) fprintf(outfd,"%d:",key[i].irsig[j]);
        fprintf(outfd," ");
        for(j=0; j <= 2 ; j++) fprintf(outfd,"%x:",key[i].nml_scode[j]);
        fprintf(outfd," ");
        for(j=0; j <= 2 ; j++) fprintf(outfd,"%x:",key[i].lfn_scode[j]);
        fprintf(outfd," ");
        for(j=0; j <= 2 ; j++) fprintf(outfd,"%x:",key[i].rfn_scode[j]);
        fprintf(outfd," ");
        for(j=0; j <= 2 ; j++) fprintf(outfd,"%d:",key[i].nml_kcode[j]);
        fprintf(outfd," ");
        for(j=0; j <= 2 ; j++) fprintf(outfd,"%d:",key[i].lfn_kcode[j]);
        fprintf(outfd," ");
        for(j=0; j <= 2 ; j++) fprintf(outfd,"%d:",key[i].rfn_kcode[j]);
        fprintf(outfd,"\n");
    } 
}

int get_keyserial(void) {
    int i = 0;
    int num = 104;               /* because 104 is not matched with any key */  
    char dmy[80];
    char str[MAX_NAME]="";

    while (1) {
        for (i = 0; i <= 102; i = i+2) {
	    sscanf(key[i].pwkname,"%[A-Z]",str);
            printf("%d:%s\n", i/2, str);
	    if ( ((i+1)%11 == 0) || (i == 102) ) {
	        printf("Type the number of the key you want.\n");
	        printf("(Enter : see next keys, x = abort) : ");
	        fgets(dmy, sizeof(dmy), stdin);
	        if (sscanf(dmy,"%d",&num) == 0) {
		    printf("Key testing aborted.\n");
		    exit(0);
		} else if ( (num >= 0) && (num <= 51) ) return num*2 ;
	    }
	}
    }
}

int get_iternum(void) {
    char dmy[80];
    int num = 30;
    while (1) {
        printf("How many times to test for a key?(20 - 200) ");
        fgets(dmy, sizeof(dmy), stdin);
        sscanf(dmy,"%d", &num);
        if ( (num >= 20) && (num <= 200) ) return num*2;
    }	
}

void get_last_idx(int *mk, int *bk) {
    int i,j;
    char dmy[80];
    int idx1=0, idx2=0, frqn_max=0, frqn_2nd_max=0;
    int stack[MAX_IRSIG];

    reset_irsig(stack);

    for (i = 0; i <= iter; i++) {
	for (j=0; irdata[i][j] != 0; j++);
	stack[--j]++;
    }

    for (j = 0; j <= MAX_IRSIG-1; j++) {
	if (frqn_max < stack[j]) {
	    frqn_max = stack[j];
	    idx1 = j;
	}
    }

    for (j = 0; j <= MAX_IRSIG-1; j++) {
	if ((frqn_2nd_max < stack[j]) && ( j != idx1)) {
	    frqn_2nd_max = stack[j];
	    idx2 = j;
	}
    }

    /* If the key is REPEAT, mk_idx = bk_idx when no erroneous signal engaged.  */
    /* But, if there is any erroneous signal, we should not trust << mk_idx >>. */
    /* This phenomenon is caused from the reason that REPEAT has only one       */
    /* signal pattern.                                                          */

    *mk = min(idx1,idx2);
    *bk = max(idx1,idx2);  
                        
    printf("We receive 2 signals patterns on each stroke (except REPEAT signal.\n");
    printf("REPEAT signal has only one signal pattern.)\n\n");
    printf("Total iteration : %d\n", iter+1);
    printf("pattern #1 : ended at index %d (%d times)\n",idx1,frqn_max);
    printf("pattern #2 : ended at index %d (%d times)\n",idx2,frqn_2nd_max);
    
    printf("Errors : %d during %d iteration\n",iter+1-frqn_max-frqn_2nd_max,iter+1);
    printf("\nPress any key to continue...\n");
    fgets(dmy, sizeof(dmy), stdin);
}

int get_start_idx(void) {
    int i,j;
    
    for (i=0; i <= 102; i++) {
	for (j=0; j <= MAX_IRSIG-1; j++) {
	    if ( (key[i].irsig[j] > key[0].irsig[j] + max_err) ||
		 (key[i].irsig[j] < key[0].irsig[j] - max_err)  )  return j;
	}
    }
    return -1;
}

void statistics(int serial, int start, int idx) {
    char dmy[80];
    int n;

    int i,j, count;
    int sum[MAX_IRSIG];
    int avr[MAX_IRSIG];
    int high[MAX_IRSIG];
    int low[MAX_IRSIG];
    int maxerr[MAX_IRSIG];

    reset_irsig(sum);
    reset_irsig(avr);
    reset_irsig(high);
    reset_irsig(low);	

    count = 0;
    for ( i = start; i <= iter; i = i+2) {

        for(j=0; irdata[i][j] != 0; j++);
	j--;
	fprintf(stderr,"j = %d\n",j);
	fprintf(stderr,"idx = %d\n", idx);
	if ( j != idx ) continue;         /* Skip erroneous signal */ 

	count++;
	if ( count==1 ) {
	    for (j=0; j <= idx; j++) sum[j]=high[j]=low[j]= irdata[i][j];
	} else {
	    for ( j=0; j <= idx; j++) {
		sum[j] += irdata[i][j];
		high[j] = max(irdata[i][j], high[j]);	    
		low[j] = min(irdata[i][j], low[j]);
	    }
	}
    }
    printf("count= %d\n",count);
    for ( j=0; j<= idx; j++) {
	avr[j] = sum[j]/count;
	maxerr[j] = max(high[j]-avr[j],avr[j]-low[j]);
    }	      

    printf("\n\nAverage values of key %s :\n\n", key[serial].pwkname);
    for (j=0; j <= idx; j++) printf("%d ",avr[j]);

    printf("\n\nThe highest numbers for each value :\n\n");
    for (j=0; j <= idx; j++) printf("%d ",high[j]);

    printf("\n\nThe Lowest numbers for each values :\n\n");
    for (j = 0 ; j <= idx; j++) printf("%d ",low[j]);
    
    printf("\n\nMax errors for each value :\n\n");
    for (j = 0 ; j <= idx; j++) printf("%d ",maxerr[j]);

    printf("\n\nDo you agree with this result? (yes = 1) : ");
    fgets(dmy, sizeof(dmy), stdin);
    sscanf(dmy,"%d",&n);
    if ( n == 1) { 
	for ( j=0; j <= idx; j++) key[serial].irsig[j] = avr[j];
    } 
}    
	   
int main(int argc, char *argv[]) {

    FILE  *infd;
    char  *infile;
    FILE  *outfd;
    char  *outfile;
    FILE  *test_fd;
    int   testfile_flag = 0;

    int   num;
    char  dmy[80];       /* variables for user input */
    int   n, ch;         /* variables for user input */
    int   serial;      
    int   i, j, rpt;     
    int   mk_idx=0, bk_idx=0;

    int fd;  
    int result;
    lirc_t data;
    unsigned long mode;
    struct stat s;
    
    if (argc == 1) goto help;
    while ( (ch = getopt(argc, argv, "hsctf:o:")) != -1 ) {
	switch(ch) {
help:	case 'h' :
	    printf("This program tests infrared signal of each key and\n");
	    printf("generate pwkrc file.\n\n"); 
	    printf("usage: irdatagen [-t] [-f infile -o outfile]\n");
	    printf("       irdatagen [-t] [-c]\n\n");
	    printf(" -h :  this help message\n");
	    printf(" -s :  print hollow pwkrc file used only for keycodegen.sh\n");
	    printf("       this option is not needed for normal users.\n");
	    printf(" -c :  creat pwkrc file from scratch.\n");
	    printf(" -f <infile>    :  input pwkrc file.\n");
	    printf(" -o <outfile>   :  output pwkrc file.\n");
	    printf(" -t :  print out signal test data as files into current directory.\n");
	    printf("       With this test data and any spreadsheet program,\n");
	    printf("       you can get more precise pwkrc file.\n");
	    exit(0);
	case 's' :
	    outfd = stdout;
	    write_to_file(outfd);
	    exit(0);
	case 'c' :
	    infile = "";
	    break;
	case 'f' :
	    infile = optarg;
	    if ( (infd = fopen( infile, "r")) == NULL ) {
		perror("Error opening input file\n");
		exit(1);
	    }
	    read_from_file(infd);
	    fclose(infd);    
	    break;
	case 'o' :
	    outfile = optarg;
	    if ( ( outfd = fopen( outfile, "w") )== NULL) {
		perror("Error opening output file\n");
		exit(1);
	    }
	    break;
	case 't' :
	    testfile_flag = 1;
	    break;
	default :
	    printf("help : irdatagen -h\n");
	    exit(0);
	}
    }
    
	    
    /******************************************************/
    /* From now, modified from mode2.c in lirc-0.6.6.tar.gz */
    /******************************************************/

    system("insmod lirc_sir");

    fd=open("/dev/lirc",O_RDONLY);
    if(fd==-1)  {
    	fprintf(stderr,"lirc: error opening /dev/lirc\n");
    	exit(EXIT_FAILURE);
    };

    if ( (fstat(fd,&s)!=-1) && (S_ISFIFO(s.st_mode)) )
    {
    	/* can't do ioctls on a pipe */
    }
    else if (ioctl(fd,LIRC_GET_REC_MODE,&mode)==-1 || mode!=LIRC_MODE_MODE2)
    {
	printf("Exit from lirc.");
    	close(fd);
      	exit(EXIT_FAILURE);
    }
    
    /*******************************************************************/
    /* End of mode2.c quoting. Thank you LIRC developers.              */
    /*******************************************************************/
    
    
    printf("\nThis program tests infrared signals of each key.\n");
    printf("The generated result will be placed as a file %s\n", outfile);
    printf("Don't omit any key data, including REPEAT.\n");    
    printf("The generated file(aka pwkrc) may not work on other machines.\n\n");
    printf("Note: If you use pwkcns driver or just want to adjust some\n");
    printf("      errneous signals of current pwkrc file, you can just \n");
    printf("      use generated pwkrc file. No more work is needed.\n");
    printf("      But, if you use pwkxwin driver, you will probably need\n");
    printf("      to run keycodegen.sh over the generated pwkrc file.\n");  
    sleep(3);

    iter = get_iternum();       
    printf("\nInput file is %s\n", infile);
    printf("Output file is %s\n", outfile);
    printf("Please, Don't omit any key.\n");
    printf("If a file with the same name of output file is, it will be deleted.\n\n");
    while (1) {

	printf("Continue test   : Enter\n");
	printf("Stop and Save   : s\n");
	printf("Stop and abort  : a\n\n");
	printf("Please select (Enter, s, a) : ");

        ch = getchar();
	fgets(dmy, sizeof(dmy), stdin);
	if ( ch  == 's' ) {
                printf("Set max_err (recommend 80 or so) :  ");
		fgets(dmy, sizeof(dmy), stdin);
		sscanf(dmy,"%d",&n);
		if ( n >= 0 ) max_err = n;

		if ( (start_idx = get_start_idx() ) == -1 ) {
		    printf("Tested data have a problem.\n");
		    printf("All keys are effectively same signal numbers.\n");
		    printf("This situation is maybe caused by incorrect.\n");
                    printf("key testing or too big max_err value.\n");
		    printf("Please check those points and test again.\n");
		    exit(1);
		}

		fprintf(stderr, "start_idx = %d\n", start_idx);
		write_to_file(outfd);
		fclose(outfd);
		exit(0);
	} else if ( ch == 'a') exit(0);


	for(i=0; i <= 219; i++) reset_irsig(irdata[i]);
	for (i=0; i <= 102; i++) {
	    if ( key[i].irsig[0] == 0 ) { 
		printf("Key %s has no IR signal data.\n", key[i].pwkname);
	    }
	}
        serial = get_keyserial();  /* We get a number of doubled real key serial */ 
 
        printf("Signals before #0 are just testing. They will not be counted.\n");
	printf("Keep repeating to press and release a key except REPEAT\n"); 
        printf("on the Palm Wireless keyboard.\n");
	printf("If you selected REPEAT key, press R key and keep it pressed).\n\n");

	rpt = -6; 	 /* iteration counter*/  
	j = 0;
	reset_irsig(irsig);

	while( (result=read(fd,&data,sizeof(data)))== sizeof(data) ) {
	    num = (int) (data&PULSE_MASK);
	    printf("%d ",num);
	    if (num < 10000) {
		if ( rpt >= 0) irdata[rpt][j] = num;
		j++;
		continue;
	    }
	    printf("\n\nIteration #%d. Go on typing.\n\n",rpt);
	    rpt++;
	    j = 0;
	    if (rpt > iter) break;
	} /* We got data for a key */

	/* print out signal test files */
	if ( testfile_flag == 1 ) {
	    char tempstr[MAX_NAME];
	    int newlength;
	    
	    /* Make proper testing filename. For example, DELETE_mk to DELETE */ 
	    newlength = strlen(key[serial].pwkname) - 3;

	    if (serial == 0) { 
		strcpy( tempstr, "REPEAT"); 
	    } else {
		strncpy( tempstr, key[serial].pwkname, newlength );
		tempstr[newlength] = '\0';
	    }

	    if ( ( test_fd = fopen( tempstr, "w")) == NULL ) {
		perror("Error opening testfile\n");
		exit(1);
	    } 
	    
	    fprintf(test_fd, "# %s\n", tempstr);
	    for (i = 0; i <= rpt - 1; i++) { 
		fprintf(test_fd, "%d : ", i);
		for (j = 0; irdata[i][j] != 0; j++) {
		    fprintf(test_fd, "%d ", irdata[i][j]);
		}
		fprintf(test_fd, "\n");
	    }
	    fclose(test_fd);
	}

	get_last_idx(&mk_idx, &bk_idx);
	fprintf(stderr,"make_idx = %d, break_idx = %d \n", mk_idx, bk_idx);


	/* Don't forget this "serial" means a key's break index. */
	/* This "serial" is even number.                         */

	for (j=0; irdata[0][j] != 0; j++);
	j--;
	if (serial==0) {
	    if ( j == bk_idx ) statistics(0, 0, bk_idx);
	} else if ( j == mk_idx ) {               /* irdata[0] is make signal  */
	    statistics(serial-1, 0, mk_idx);        
 	    statistics(serial, 1, bk_idx);    
	} else if ( j == bk_idx ) {               /* irdata[0] is break signal  */
	    statistics(serial, 0, bk_idx);    
	    statistics(serial-1, 1, mk_idx);        
	} else {
	    printf("Sorry. Unfortunately, Test #0 has a bad signal.\n");
	    printf("Pleae test again this key. \n");
	}
		    
    } /* close while(1) */ 
}

