/****************************************************************************
 ****************************************************************************
 * Palm Wireless Keyboard(TM) add-on driver for linux-console. v0.2.2
 * Copyright GPL Licence. Quoted some part of the source code from
 * mode2.c in the lirc-0.6.6.tar.gz package. Its Copyright is bellow.
 * Copyright (C) 2004.12.30 lampo1 <lampo1@paran.com>
 *
 ****************************************************************************
 * mode2 - shows the pulse/space length of a remote button
 *
 * Copyright (C) 1998 Trent Piepho <xyzzy@u.washington.edu>
 * Copyright (C) 1998 Christoph Bartelmus <lirc@bartelmus.de>
 *
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <fcntl.h>
#include <signal.h>
#include <termios.h>

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include "lirc.h"

#define MAX_NAME   14
#define MAX_IRSIG  60
#define MAX_SCODE  3
#define TYPERATE   5      

struct pwkey {
    int    serial;
    char   pwkname[MAX_NAME];
    char   nml_xname[MAX_NAME];
    int    irsig[MAX_IRSIG];
    int    nml_scode[MAX_SCODE];
    int    lfn_scode[MAX_SCODE];
    int    rfn_scode[MAX_SCODE];
};

struct pwkey key[103]; 

int max_err = 80;           /* limit of acceptable difference between signals */
int start_idx = 0;          /* starting index of comparing */ 
int irsig[MAX_IRSIG];

void reset_irsig(int *irsig) {
    int i;
    for (i = 0; i <= MAX_IRSIG-1; i++) irsig[i] = 0;
}

void read_from_file(FILE *infd) {
    int i, j;
    char dmy[80];
    
    do {
	fgets(dmy, sizeof(dmy), infd);        
    } while( dmy[0] == '#' );

    sscanf(dmy, "%*s %*s %d %*s %*s %*s %d",&max_err, &start_idx);
    for (i = 0; i<= 102; i++) {
	fscanf(infd,"%d %s %s %*s %*s", &key[i].serial, key[i].pwkname,
	       key[i].nml_xname);
	for(j=0; j <= MAX_IRSIG-1; j++) fscanf(infd,"%d:",&key[i].irsig[j]);
        for(j=0; j <= 2 ; j++) fscanf(infd,"%x:",&key[i].nml_scode[j]);
        for(j=0; j <= 2 ; j++) fscanf(infd,"%x:",&key[i].lfn_scode[j]);
        for(j=0; j <= 2 ; j++) fscanf(infd,"%x:",&key[i].rfn_scode[j]);
        for(j=0; j <= 2 ; j++) fscanf(infd,"%*d:");
        for(j=0; j <= 2 ; j++) fscanf(infd,"%*d:");
        for(j=0; j <= 2 ; j++) fscanf(infd,"%*d:");

	fgets(dmy, sizeof(dmy), infd);   
    }

    #ifdef DEBUG
    printf("max_err = %d, start_idx = %d\n", max_err, start_idx);   
    for (i = 0; i <= 102; i++) {
	printf("%d %s %s ", i, key[i].pwkname, key[i].nml_xname);
	for(j=0; j <= MAX_IRSIG-1; j++) printf("%d:",key[i].irsig[j]);
        printf(" ");
        for(j=0; j <= 2 ; j++) printf("%d:",key[i].nml_scode[j]);
        printf(" ");
        for(j=0; j <= 2 ; j++) printf("%d:",key[i].lfn_scode[j]);
        printf(" ");
        for(j=0; j <= 2; j++)  printf("%d:",key[i].rfn_scode[j]);
        printf("\n");
    }
    #endif    
}

int find_key(int idx, int num) {

    int i;
    int matched_keys;    /* number of matched keys */
    int key_found;
    static int key_to_check[103];   /* if key_to_check[i] = 1, check the key[i] */ 

    if (idx == start_idx) {
	for (i = 0; i <= 102 ; i++) key_to_check[i] = 1;       
    }   /* all keys are to be checked. */

    matched_keys = 0;
    for ( i = 0; i <= 102; i++) {
	if (key_to_check[i]==1) {

	  /*    #ifdef DEBUG
	    fprintf(stderr,"Checking key[%d]...\n",i);
	    fprintf(stderr,"key[%d].irsig[%d]=%d\n",i,idx,key[i].irsig[idx]); 
	    #endif */
   	 
	    if ( ( key[i].irsig[idx] <= num + max_err) && 
   	         ( key[i].irsig[idx] >= num - max_err) ) {
                key_to_check[i] = 1;
		matched_keys++;
		key_found = i;
            } else key_to_check[i] = 0;
	}
    }

    if (matched_keys == 0) return -1;         /* no key has matched */
    else if (matched_keys >= 2) return 103;   /* more than 2 keys are matching */
    return key_found;   
} 
        
int main() {

    FILE *infd;
    char *infile = "/etc/pwkrc";
    FILE *outfd;
    char *outfile = "/dev/kbde";

    sigset_t  newset;    

    int num;                    /* ir input */
    int *sc_ptr;                /* pointer to scancode */
    int idx;                    /* index of a signal array */
    int n;                      /* contain key sereal */
    int lfn_FLAG=0, rfn_FLAG=0;
    int found;                  /* indicate whether key is found or not */
    int n_REPEAT;               /* count repetition of REPEAT signals */
    int i, j;                   /* mainly i is used for key serial, j for index */
    int prev = 74;              /* initial value 74 is LSHIFT_bk key. No harm. */    
    int typerate=0;      

    #ifdef SQNCHECK
    int mk_table[103];          /* 1 or 0 */
    /* mk_table[] is used for simple input sequence checking.   */
    /* For example, mk_table[7] = 1 indicates key A had been pressed */
    /* before. You can detect unpressed key's release by this table. */
    #endif

    /******************************************************/
    /* From now, quoted from mode2.c in lirc-0.6.6.tar.gz */
    /******************************************************/
   
    int fd;  
    int result;
    lirc_t data;
    unsigned long mode;
    struct stat s;
    
    system("insmod lirc_sir");

    fd=open("/dev/lirc",O_RDONLY);
    if (fd==-1)  {
    	fprintf(stderr,"Error opening /dev/lirc\n");
    	exit(EXIT_FAILURE);
    };
    
    if ( (fstat(fd,&s)!=-1) && (S_ISFIFO(s.st_mode)) )
    {
    	/* can't do ioctls on a pipe */
    }
    else if(ioctl(fd,LIRC_GET_REC_MODE,&mode)==-1 || mode!=LIRC_MODE_MODE2)
    {
    	printf("This program is only intended for receivers "
    	       "supporting the pulse/space layer.\n");
    	printf("Note that this is no error, but this program simply "
    	       "makes no sense for your\n"
    	       "receiver.\n");
    	close(fd);
      	exit(EXIT_FAILURE);
    }

    /*******************************************************************/
    /* End of mode2.c quoting. Thank you LIRC developers.              */
    /*******************************************************************/

    /* Blocking all signals. */
    sigfillset(&newset);
    sigprocmask(SIG_BLOCK, &newset, NULL);  

    /* get ir data from pwkrc file */
    if ( (infd = fopen( infile, "r"))== NULL ) {
	perror("Error opening datafile\n");
	exit(1);
    }
    read_from_file(infd);          
    fclose(infd);
    
    /* open output device */
    if ( (outfd = fopen(outfile, "w"))== NULL ) {
	perror("Error opening kbde device file.\n");
	exit(1);
    }

    idx = 0; 
    found = 0;
    while( (result=read(fd,&data,sizeof(data)))== sizeof(data) ) {

        num = (int) (data&PULSE_MASK);            /* get ir signal */
 
        if ((num >= 10000) || (idx >= MAX_IRSIG)) {
	    idx = 0;
	    reset_irsig(irsig);
	    found = 0;
            continue;
        }              
	    	    
        if (found == 1) continue;       /* no more need to find */
	else irsig[idx] = num;   

        if ( idx < start_idx ) { 
            idx++;
            continue; 
        };                       
 
	n = find_key(idx,num);

	if ( n == 103 ) {  /* return value 103 means 2 or more keys */
			   /* are matching until now.               */ 
	    idx++;
	    continue;
	}

	found = 1;

        if ( n == -1) {    /* no key has matched in the end. */

            #ifdef DEBUG
            fprintf(stderr,"n = -1, Bad signal\n");
            for ( j=0; j <= MAX_IRSIG-1; j++) fprintf(stderr, "%d ",irsig[j]);
            fprintf(stderr,"\n");
            #endif

	    reset_irsig(irsig);  /* Bad signal */
	    continue;                     
        }

	if ( n == 0 ) {       /* REPEAT */

            #ifdef DEBUG
	          fprintf(stderr,"%s ",key[n].pwkname);
	          fprintf(stderr,"n_REPEAT=%d ",n_REPEAT);
	    #endif

	    #ifdef SQNCHECK
	    if ( (prev%2) == 0 ) {

	        #if defined(SQNCHECK) && defined(DEBUG)
                fprintf(stderr, "n = 0, prev = %d, irrational\n", prev);
                #endif
		    
                continue;     /* something wrong. break keystroke cannot repeat. */
            }
	    #endif

            n_REPEAT++;
            if (n_REPEAT >= 10) {   /* ten times of same signal generates key repeating */   
                typerate = TYPERATE;
                if (n_REPEAT == UINT_MAX) n_REPEAT = 0;
                n = prev;
            } else {
		continue;
            }
	} else n_REPEAT = 0;

	#ifdef SQNCHECK
	if ( (n % 2) == 1 ) mk_table[n] = 1;
        else if ( mk_table[n-1] == 0 ) {
	    
	    #if defined(SQNCHECK) && defined(DEBUG)
            fprintf(stderr,"n = %d, mk_table[%d-1] = 0, irrational\n",  n, n);
	    fprintf(stderr,"It means that unpressed key is released.\n");
	    fprintf(stderr,"Maybe somthing wrong situation.\n");
	    #endif

            continue;   /* something wrong. Consider it as a bad signal */
        } else mk_table[n-1] = 0;
	#endif

	switch (n) {
        case 1  : lfn_FLAG = 1; break;  
        case 2  : lfn_FLAG = 0; break;  
        case 3  : rfn_FLAG = 1; break;
        case 4  : rfn_FLAG = 0; break;

        default : if (lfn_FLAG == 1) sc_ptr = key[n].lfn_scode; 
                  else if (rfn_FLAG == 1) sc_ptr = key[n].rfn_scode; 
	          else  sc_ptr = key[n].nml_scode; 

                  #ifdef DEBUG 
                  fprintf(stderr,"lfn_FLAG=%d,rfn_flag=%d\n",lfn_FLAG,rfn_FLAG);
    	          fprintf(stderr,"*sc_ptr = %d %d\n", sc_ptr[0], sc_ptr[1]);
	          #endif

        	  for (j=0; j <= typerate; j++) {
                      for (i=0; sc_ptr[i] != 0 ; i++) { 
			  putc(sc_ptr[i], outfd);
			  fflush(outfd);
		      }
		  }
	          prev = n;
	          typerate = 0;
	}  /* close switch.  We sent scancodes for the key.  */  
    } /* close while. */
}  

